""" base_class consists of the building blocks of models: the base classes models should extend from and relevant utility functions.
"""
from __future__ import annotations

import inspect
import io
import logging
import os
import sys
from abc import abstractmethod
from typing import Any, Callable, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.metrics import (
    average_precision_score,
    explained_variance_score,
    max_error,
    mean_absolute_error,
    mean_squared_error,
    r2_score,
    roc_auc_score,
)

import olorenchemengine as oce
from olorenchemengine.internal import *

# List of metrics that can be used for evaluating classification models
classification_metrics = {
    "Average Precision": average_precision_score,
    "ROC-AUC": roc_auc_score,
}

# List of metrics that can be used for evaluating regression models
from scipy.stats import spearmanr

regression_metrics = {
    "r2": r2_score,
    "Spearman": lambda y, y_pred: spearmanr(y, y_pred)[0],
    "Explained Variance": explained_variance_score,
    "Max Error": max_error,
    "Mean Absolute Error": mean_absolute_error,
    "Mean Squared Error": mean_squared_error,
    "Root Mean Squared Error": lambda y, y_pred: np.sqrt(mean_squared_error(y, y_pred)),
}

# Entire list of metrics that can be used for evaluation models
metric_functions = {**classification_metrics, **regression_metrics}

# Dictionary indicating whether a better model has a 'higher' metric value or a 'lower' metric value for a given materic
metric_direction = {
    "Average Precision": "higher",
    "ROC-AUC": "higher",
    "r2": "higher",
    "Explained Variance": "higher",
    "Max Error": "lower",
    "Mean Absolute Error": "lower",
    "Mean Squared Error": "lower",
    "Root Mean Squared Error": "lower",
    "Spearman": "higher",
}


class BaseObject(BaseClass):
    """BaseObject is the parent class for all classes which directly wrap some object to be saved via joblib.

    Attributes:
        obj (object): the object which is wrapped by the BaseObject
    """

    @log_arguments
    def __init__(self, obj=None):
        self.obj = obj

    def _save(self):
        import joblib

        b = io.BytesIO()
        joblib.dump(self.obj, b)
        return {"obj": b.getvalue()}

    def _load(self, d):
        import joblib

        super()._load(d)
        self.obj = joblib.load(io.BytesIO(d["obj"]))


class BaseEstimator(BaseObject):

    """Utility class used to wrap any object with a fit and predict method"""

    def fit(self, X, y):
        """Fit the estimator to the data

        Parameters:
            X (np.array): The data to fit the estimator to
            y (np.array): The target data to fit the estimator to
        Returns:
            self (object): The estimator object fit to the data
        """

        return self.obj.fit(X, y)

    def predict(self, X):
        """Predict the output of the estimator

        Parameters:
            X (np.array): The data to predict the output of the estimator on
        Returns:
            y (np.array): The predicted output of the estimator
        """
        pred = self.obj.predict(X)
        pred = np.array(pred)
        if pred.ndim == 1:
            pred = pred
        else:
            pred = np.array([pred])
        return pred


class LinearRegression(BaseEstimator):

    """Wrapper for sklearn LinearRegression"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.linear_model import LinearRegression as LinearRegression_

        self.obj = LinearRegression_(*args, **kwargs)


class BasePreprocessor(BaseObject):
    """BasePreprocessor is the parent class for all preprocessors which transform the features or properties of a dataset.

    Methods:
        fit: fit the preprocessor to the dataset
        fit_transform: fit the preprocessor to the dataset return the transformed values
        transform: return the transformed values
        inverse_transform: return the original values from the transformed values
    """

    def fit(self, X):
        """Fits the preprocessor to the dataset.

        Parameters:
            X (np.ndarray): the dataset

        Returns:
            The fit preprocessor instance
        """
        X = np.array(X)
        if len(X.shape) == 1:
            X = X.reshape(-1, 1)
        return self.obj.fit(X)

    def fit_transform(self, X):
        """Fits the preprocessor to the dataset and returns the transformed values.

        Parameters:
            X (np.ndarray): the dataset

        Returns:
            The transformed values of the dataset as a numpy array
        """
        X = np.array(X)
        if len(X.shape) == 1:
            X = X.reshape(-1, 1)
        X = self.obj.fit_transform(X)
        return X

    def transform(self, X):
        """Returns the transformed values of the dataset as a numpy array.

        Parameters:
            X (np.ndarray): the dataset

        Returns:
            The transformed values of the dataset as a numpy array
        """
        X = np.array(X)
        if len(X.shape) == 1:
            X = X.reshape(-1, 1)
        X = self.obj.transform(X)
        return X

    def inverse_transform(self, X):
        """Returns the original values from the transformed values.

        Parameters:
            X (np.ndarray): the transformed values

        Returns:
            The original values from the transformed values
        """
        if len(X.shape) == 1:
            X = X.reshape(-1, 1)
        X = np.array(X)
        return self.obj.inverse_transform(X)


class QuantileTransformer(BasePreprocessor):
    """QuantileTransformer is a BasePreprocessor which transforms a dataset by quantile transformation to specified distribution.

    Attributes:
        obj (sklearn.preprocessing.QuantileTransformer): the object which is wrapped by the BasePreprocessor
    """

    @log_arguments
    def __init__(
        self,
        n_quantiles=1000,
        output_distribution="normal",
        subsample=1e5,
        random_state=None,
    ):

        from sklearn.preprocessing import QuantileTransformer

        self.obj = QuantileTransformer(
            n_quantiles=n_quantiles,
            output_distribution=output_distribution,
            subsample=subsample,
            random_state=random_state,
        )


class StandardScaler(BasePreprocessor):
    """StandardScaler is a BasePreprocessor which standardizes the data by removing the mean and scaling to unit variance.

    Attributes:
        obj (sklearn.preprocessing.StandardScaler): the object which is wrapped by the BasePreprocessor
    """

    @log_arguments
    def __init__(self, with_mean=True, with_std=True):

        from sklearn.preprocessing import StandardScaler

        self.obj = StandardScaler(with_mean=with_mean, with_std=with_std)


class LogScaler(BasePreprocessor):
    """LogScaler is a BasePreprocessor which standardizes the data by taking the log and then removing the mean and scaling to unit variance."""

    @log_arguments
    def __init__(self, with_mean=True, with_std=True):

        from sklearn.preprocessing import StandardScaler

        self.obj = StandardScaler(with_mean=with_mean, with_std=with_std)

    def fit(self, X):
        """Fits the preprocessor to the dataset.

        Parameters:
            X (np.ndarray): the dataset

        Returns:
            The fit preprocessor instance
        """
        X = np.array(X)
        self.mean = np.mean(X)
        X = np.log10(X + self.mean + 1e-3)
        return self.obj.fit(X.reshape(-1, 1))

    def fit_transform(self, X):
        """Fits the preprocessor to the dataset and returns the transformed values.

        Parameters:
            X (np.ndarray): the dataset

        Returns:
            The transformed values of the dataset as a numpy array
        """
        X = np.array(X)
        self.mean = np.mean(X)
        X = np.log10(X + self.mean + 1e-3)
        return self.obj.fit_transform(X.reshape(-1, 1)).reshape(-1)

    def transform(self, X):
        """Returns the transformed values of the dataset as a numpy array.

        Parameters:
            X (np.ndarray): the dataset

        Returns:
            The transformed values of the dataset as a numpy array
        """
        X = np.log10(np.array(X) + self.mean + 1e-3)
        return self.obj.transform(X.reshape(-1, 1)).reshape(-1)

    def inverse_transform(self, X):
        """Returns the original values from the transformed values.

        Parameters:
            X (np.ndarray): the transformed values

        Returns:
            The original values from the transformed values
        """
        X = np.array(X)
        return (
            10 ** (self.obj.inverse_transform(X.reshape(-1, 1)).reshape(-1))
            - self.mean
            - 1e-3
        )

    def _save(self):
        d = super()._save()
        d["mean"] = self.mean
        return d

    def _load(self, d):
        super()._load(d)
        self.mean = d["mean"]
        return self


def detect_setting(data):
    values, _ = np.unique(data, return_counts=True)
    if len(values) <= 2:
        return "classification"
    else:
        return "regression"


class BaseModel(BaseClass):
    """
    BaseModel for training and evaluating different models

    Parameters:
        normalization (BasePreprocessor or str): the normalization to be used for the data
        setting (str): whether the model is a "classification" model or a "regression" model.
            Default is "auto" which automatically detects the setting from the dataset.
        model_name (str): the name of the model. Default is None, which instructs
            BaseModel to use `model_name_from_model` to select the name of the model.

    Attributes:
        setting (str): whether the model is a "classification" model or a "regression" model
        model_name (str): the name of the model, either assigned or a hashed name

    Methods:
        preprocess: preprocess the inputted data into the appropriate format
        _fit: fit the model to the preprocessed data, to be used internally implemented by child classes
        fit: fit the model to the inputted data, user can specify if they want regression or classification using the setting parameter.
        _predict: predict the properties of the inputted data, to be used internally implemented by child classes
        predict: predict the properties of the inputted data
        test: test the model on the inputted data, output metrics and optionally predicted values
        copy: returns a copy of the model (internal state not copied)
    """

    def __init__(self, normalization="zscore", setting="auto", name=None, **kwargs):

        self.normalization = normalization
        self.setting = setting
        if name is None:
            self.name = model_name_from_model(self)
        else:
            self.name = name

        self.calibrator = None
        self.em_status = None

    def preprocess(self, X, y, fit=False):
        """

        Args:
            X (list of smiles)

        Returns:
            Processed list converted into whatever input for the model
        """
        return X

    def visualize_parameters_ipynb(self):
        from urllib.parse import quote

        from IPython.display import IFrame, display

        parameters = parameterize(self)
        display(
            IFrame(
                f"https://oas.oloren.ai/jsonvis?json={quote(json.dumps(parameters))}",
                width=800,
                height=500,
            )
        )

    @abstractmethod
    def _fit(self, X_train, y_train: np.ndarray) -> None:
        """To be implemented by the child class; fits the model on the provided dataset given preprocessed features.

         Note: _fit shouldn't be directly called by the user, rather it should be called indirectly via the fit method.

        Args:
            X_train: preprocessed features
            y_train (np.ndarray): values to predict
        """
        pass

    def fit(
        self,
        X_train: Union[pd.DataFrame, np.ndarray],
        y_train: Union[pd.Series, list, np.ndarray],
        valid: Tuple[
            Union[pd.DataFrame, np.ndarray], Union[pd.Series, list, np.ndarray]
        ] = None,
        error_model: BaseErrorModel = None,
    ):
        """Calls the _fit method of the model to fit the model on the provided dataset.

        Args:
            X_train (Union[pd.DataFrame, np.ndarray]): Input data to be fit on (structures + optionally features) e.g. a pd.DataFrame containing a "smiles" column or a list of experimental data
            y_train (Union[pd.Series, list, np.ndarray]): Values to predict from the input data
            valid (Tuple[Union[pd.DataFrame, np.ndarray], Union[pd.Series, list, np.ndarray]]): Optional validation data, which can be used as with methods like early stopping and model averaging.
            error_model (BaseErrorModel): Optional error model, which can be used to predict confidence intervals.
        """
        y_train = np.array(y_train)

        X_train_original = X_train
        y_train_original = y_train

        if not valid is None:
            X_valid, y_valid = valid

        # Pick setting and normalize y-values as appropriate
        if self.setting == "auto":
            self.setting = detect_setting(y_train)
        if self.setting == "classification":
            y_train = np.array(y_train).astype(int)
            if not valid is None:
                y_valid = np.array(y_valid).astype(int)
            logging.info("Classification dataset detected")
        elif self.setting == "regression":
            if self.normalization == "zscore":
                self.ymean = np.mean(y_train)
                self.ystd = np.std(y_train)
                y_train = (y_train - self.ymean) / self.ystd
                if not valid is None:
                    y_valid = (y_valid - self.ymean) / self.ystd
            elif issubclass(type(self.normalization), BasePreprocessor):
                self.normalization = self.normalization.copy()
                y_train = self.normalization.fit_transform(y_train)
                y_train = np.array(y_train)
                if not valid is None:
                    y_valid = self.normalization.transform(y_valid)
                    y_valid = np.array(y_valid)
            if (np.isreal(y_train.any()) == False) & (str(y_train).find("%") == -1):
                logging.warning(
                    " Dataset contains non-numeric values, which is incompatible with regression."
                )
                y_train.str.replace(r"[^0-9.%]", "", regex=True)
            logging.info("Regression dataset detected")
        else:
            logging.error(
                "Setting was not defined as either auto, regression, or classification."
            )
            sys.exit(1)

        # Train model
        X_train = self.preprocess(X_train, y_train, fit=True)
        if not valid is None:
            X_valid = self.preprocess(X_valid, y_valid, fit=False)

        self._fit(X_train, y_train)

        # Calibrate model
        if not valid is None and self.setting == "classification":
            self.calibrate_model(X_valid, y_valid)

        # Build error model
        if not error_model is None:
            error_model.build(self, X_train_original, y_train_original)
            self.error_model = error_model
            self.em_status = "built"

    def calibrate(self, X_valid, y_valid):
        y_pred_valid = self.predict(X_valid)

        from sklearn.calibration import calibration_curve

        prob_pred, prob_true = calibration_curve(
            y_valid, y_pred_valid, n_bins=min(len(y_valid), 10)
        )
        prob_pred, prob_true = zip(
            *[x for x in zip(prob_pred, prob_true) if 0 < x[1] < 1]
        )
        prob_true = np.array(prob_true)
        prob_true = np.log(prob_true / (1 - prob_true))
        prob_pred = np.array(prob_pred)
        if np.unique(prob_true).shape[0] > 1 and np.unique(prob_pred).shape[0] > 1:
            self.calibrator = LinearRegression()
            self.calibrator.fit(prob_pred.reshape(-1, 1), prob_true.reshape(-1, 1))

    def fit_class(
        self,
        X_train: Union[pd.DataFrame, np.ndarray],
        y_train: Union[pd.Series, list, np.ndarray],
        valid: Tuple[
            Union[pd.DataFrame, np.ndarray], Union[pd.Series, list, np.ndarray]
        ] = None,
    ):
        from sklearn.model_selection import KFold
        from tqdm import tqdm

        kf = KFold(n_splits=5)
        y_pred = list()
        y_true = list()
        for train, test in tqdm(kf.split(X_train)):
            X_fold_train = X_train.loc[train]
            y_fold_train = y_train.loc[train]
            X_fold_test = X_train.loc[test]
            y_fold_test = y_train.loc[test]
            self.fit(X_fold_train, y_fold_train)
            preds = self.predict(X_fold_test)
            if not isinstance(preds, list):
                preds = preds.tolist()
            y_pred += preds
            y_true += y_fold_test.tolist()

        from sklearn.calibration import calibration_curve

        prob_pred, prob_true = calibration_curve(
            y_true, y_pred, n_bins=min(len(y_true), 10)
        )
        prob_pred, prob_true = zip(
            *[x for x in zip(prob_pred, prob_true) if 0 < x[1] < 1]
        )
        prob_true = np.array(prob_true)
        prob_true = np.log(prob_true / (1 - prob_true))
        prob_pred = np.array(prob_pred)
        self.calibrator = LinearRegression()
        self.calibrator.fit(prob_pred.reshape(-1, 1), prob_true.reshape(-1, 1))

        self.fit(X_train, y_train)

    def _unnormalize(self, Y):
        if (
            self.normalization == "zscore"
            and hasattr(self, "ymean")
            and hasattr(self, "ystd")
        ):
            result = Y * self.ystd + self.ymean
        elif issubclass(type(self.normalization), BasePreprocessor):
            result = self.normalization.inverse_transform(Y)
        else:
            result = Y
        return result

    @abstractmethod
    def _predict(self, X: Union[pd.DataFrame, np.ndarray]) -> np.ndarray:
        """To be implemented by the child class; returns the predicted values for provided dataset given preprocessed features.

        Note: _predict shouldn't be directly called by the user, rather it should be called indirectly via the predict method.

        Args:
            X (Union[pd.DataFrame, np.ndarray]): Input data to be predicted (structures + optionally features), will be preprocessed by the preprocess method.

        Returns:
            np.ndarray: Predicted values for the provided dataset.
            Shape: (number of samples, number of predicted values)
        """
        pass

    def predict(
        self,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
        return_ci=False,
        return_vis=False,
        skip_preprocess=False,
        **kwargs,
    ) -> np.ndarray:
        """Calls the _predict method of the model and returns the predicted values for provided dataset.

        Args:
            X (Union[pd.DataFrame, np.ndarray, list, pd.Series]): Input data to be predicted (structures + optionally features), will be preprocessed by the preprocess method.
            return_ci (bool): If error model fitted, whether or not to return the confidence intervals.
            return_vis (bool): If error model fitted, whether or not to return BaseVisualization objects.

        Returns:
            If return_ci and return_vis are False:
                np.ndarray: Predicted values for the provided dataset.
                Shape: (number of samples, number of predicted values)
            If return_ci or return_vis are true:
                pd.DataFrame: Predicted values, confidence intervals, and/or error plots for the provided dataset.
        """
        X_original = X
        if not skip_preprocess:
            X = self.preprocess(X, None, fit=False)
        Y = self._predict(X, **kwargs)
        if self.setting == "regression":
            result = self._unnormalize(Y)
            if self.calibrator is not None:
                result = self.calibrator.predict(result.reshape(-1, 1)).reshape(-1)
            if return_ci or return_vis:
                assert self.em_status == "fitted", "error model not fit"
                result = pd.DataFrame({"predicted": result})
                errors = self.error_model.score(X_original)
                if return_ci:
                    result["ci"] = errors
                if return_vis:
                    from olorenchemengine.visualizations.visualization import (
                        VisualizeError,
                    )

                    ci = self.error_model.quantile
                    result["vis"] = [
                        VisualizeError(
                            self.error_model.y_train,
                            float(result["predicted"][i]),
                            float(errors[i]),
                            ci=ci,
                        )
                        for i in range(len(errors))
                    ]
            return result
        else:
            if self.calibrator is not None:
                Y = np.array(Y)
                Y = Y.reshape(-1, 1)
                Y = self.calibrator.predict(Y)
                return 1 / (1 + np.exp(-Y.flatten()))
            else:
                return Y

    def fit_cv(
        self,
        X: Union[pd.DataFrame, np.ndarray],
        y: Union[pd.Series, list, np.ndarray],
        n_splits: int = 5,
        error_model: BaseErrorModel = None,
        ci: float = 0.8,
        scoring: str = None,
    ):
        """Trains a production-ready model.

        This method trains the model on the entire dataset. It also performs an
        intermediate cross-validation step over dataset to both generate test
        metrics across the entire dataset, as well as to generate information
        which is used to calibrate the trained model.

        Calibration means to ensure that the probabilities outputted by
        classifiers reflect true distributions and to create appropriate
        confidence intervals for regression data.

        Args:
            X_train (Union[pd.DataFrame, np.ndarray]): Input data to be fit on (structures + optionally features) e.g. a pd.DataFrame containing a "smiles" column or a list of experimental data
            y_train (Union[pd.Series, list, np.ndarray]): Values to predict from the input data
            n_splits (int): Number of cross validation splits, default 5
            error_model (BaseErrorModel): Optional error model, which can be used to predict confidence intervals.
            ci (float): the confidence interval predicted by the error model
            scoring (str): Metric function to use for scoring cross validation splits; must be in `metric_functions`

        Returns:
            list: Cross validation metrics for each split
        """

        self.fit(X, y, error_model=error_model)

        residuals = None
        scores = None
        pred = None
        true = None

        cross_val_metrics = []
        X = np.array(X).flatten()
        y = np.array(y).flatten()

        if scoring is None:
            if self.setting == "regression":
                scoring = "Explained Variance"
            elif self.setting == "classification":
                scoring = "Average Precision"

        from sklearn.calibration import calibration_curve
        from sklearn.model_selection import KFold

        kf = KFold(n_splits=n_splits)

        for train_index, test_index in kf.split(X):
            X_train, X_test = X[train_index], X[test_index]
            y_train, y_test = y[train_index], y[test_index]
            model = self.copy()
            model.fit(X_train, y_train)
            y_pred_test = model.predict(X_test)

            if hasattr(self, "error_model") and self.setting == "regression":
                y_pred_test = np.array(y_pred_test).flatten()
                error_model = type(self.error_model)(
                    *self.error_model.args, **self.error_model.kwargs
                )
                error_model.build(model, X_train, y_train)
                if isinstance(error_model, BaseAggregateErrorModel):
                    error_model.train_cv()
                new_scores = error_model.calculate(X_test, y_pred_test)
                if scores is None:
                    pred = y_pred_test
                    true = y_test
                    scores = new_scores
                else:
                    scores = np.concatenate((scores, new_scores))
                    pred = np.concatenate((pred, y_pred_test))
                    true = np.concatenate((true, y_test))
            elif self.setting == "classification":
                prob_pred, prob_true = calibration_curve(
                    y_test, y_pred_test, n_bins=min(len(y_test), 10)
                )
                prob_pred, prob_true = zip(
                    *[x for x in zip(prob_pred, prob_true) if 0 < x[1] < 1]
                )
                prob_true = np.array(prob_true)
                prob_true = np.log(prob_true / (1 - prob_true))
                prob_pred = np.array(prob_pred)
                if pred is None:
                    pred = prob_pred
                    true = prob_true
                else:
                    pred = np.concatenate((pred, prob_pred))
                    true = np.concatenate((true, prob_true))

            cross_val_metrics.append(metric_functions[scoring](y_test, y_pred_test))

        if self.setting == "regression":
            self.calibrator = LinearRegression()
            self.calibrator.fit(pred.reshape(-1, 1), true.reshape(-1, 1))
            pred = self.calibrator.predict(pred.reshape(-1, 1)).reshape(-1)
            true = true.reshape(-1)
            residuals = np.abs(pred - true)
            if hasattr(self, "error_model"):
                self.error_model._fit(residuals, scores, quantile=ci)
                self.em_status = "fitted"
        elif self.setting == "classification":
            self.calibrator = LinearRegression()
            self.calibrator.fit(pred.reshape(-1, 1), true.reshape(-1, 1))

        return cross_val_metrics

    def test(
        self,
        X: Union[pd.DataFrame, np.ndarray],
        y: Union[pd.Series, list, np.ndarray],
        values: bool = False,
        fit_error_model: bool = False,
    ) -> dict:
        """Tests the model on the provided dataset returning a dictionary of metrics and optionally the predicted values.

        Args:
            X (Union[pd.DataFrame, np.ndarray]): Input test data to be predicted (structures + optionally features)
            y (Union[pd.Series, list, np.ndarray]): True values for the properties
            values (bool, optional): Whether or not to return the predicted values for the test data. Defaults to False.
            fit_error_model (bool): If present, whether or not to fit the error model on the test data.

        Returns:
            dict: Dictionary of metrics and optionally the predicted values
        """
        if fit_error_model and hasattr(self, "error_model"):
            self.error_model.fit(X, y)
            self.em_status = "fitted"

        import json

        import joblib

        dataset_hash = joblib.hash(X) + joblib.hash(y)

        y_pred = self.predict(X)

        import pandas as pd

        if isinstance(y, pd.Series):
            y = y.tolist()

        d = {}

        if self.setting == "classification":
            d.update(
                {k: float(v(y, y_pred)) for k, v in classification_metrics.items()}
            )
        elif self.setting == "regression":
            d.update({k: float(v(y, y_pred)) for k, v in regression_metrics.items()})

        # LOGGING START
        import requests

        try:
            response = requests.get(
                f"https://api.oloren.ai/firestore/log_model_performance",
                params={
                    "name": self.name,
                    "dataset_hash": dataset_hash,
                    "metrics": json.dumps(d),
                    "params": json.dumps(parameterize(self)),
                },
            )
        except:
            pass
        # LOGGING END

        if values:
            d.update({"values": y_pred})

        return d

    def _save(self) -> dict:
        d = {}
        if hasattr(self, "ymean") and hasattr(self, "ystd"):
            d.update({"ymean": self.ymean})
            d.update({"ystd": self.ystd})
        if hasattr(self, "setting"):
            d.update({"setting": self.setting})
        if not self.calibrator is None:
            d.update({"calibrator": self.calibrator._save()})
        if hasattr(self, "error_model"):
            d.update({"error_model": saves(self.error_model)})
        if hasattr(self, "em_status"):
            d.update({"em_status": self.em_status})
        if hasattr(self, "normalization") and issubclass(
            type(self.normalization), BasePreprocessor
        ):
            d.update({"normalization": self.normalization._save()})
        return d

    def _load(self, d) -> None:
        if "ymean" in d.keys() and "ystd" in d.keys():
            self.ymean = d["ymean"]
            self.ystd = d["ystd"]
        if "setting" in d.keys():
            self.setting = d["setting"]
        if "calibrator" in d.keys():
            self.calibrator = LinearRegression()
            self.calibrator._load(d["calibrator"])
        if "error_model" in d.keys():
            self.error_model = loads(d["error_model"])
            self.error_model.model = self
        if "em_status" in d.keys():
            self.em_status = d["em_status"]
        if "normalization" in d.keys():
            self.normalization._load(d["normalization"])

    def copy(self) -> BaseModel:
        """returns a copy of itself

        Returns:
            BaseModel: copied instance of itself
        """
        if hasattr(self, "args"):
            return type(self)(*self.args, **self.kwargs)
        else:
            return type(self)()

    def upload_oas(self, fname: str = None):
        """uploads the BaseClass object to cloud for OAS access. Model must be trained.

        Args:
            fname (str) (optional): the file name to name the uploaded model file. If left empty/None, names the file with default name associated with the BaseClass object.
        """
        try:
            self.predict("C")
        except:
            raise Exception("Model for upload must be trained.")

        if not fname:
            fname = self.name

        out = oas_connector.upload_model(self, fname)
        print(f"Access your model at: https://oas.oloren.ai/endpoint/?mid={out[1].id}")


class MakeMultiClassModel(BaseModel):
    """
    Base class for extending the classification capabilities of BaseModel to more than two classes, e.g. classes {W,X,Y,Z}.
    Uses the commonly-implemented One-vs-Rest (OvR) strategy. For each classifier, the class is fitted against all the other classes. The probabilities are then normalized and compared for each class.

    Parameters:
        individual_classifier (BaseModel): Model for binary classification tasks,
            which is to be turned into a multi-class model.
    """

    @log_arguments
    def __init__(self, individual_classifier: BaseModel) -> None:
        super().__init__()
        self.classifier = individual_classifier
        self.classifiers = []

    def _fit(self, X_train, y_train: np.ndarray) -> None:
        """Empty method to correctly override the fit method of the parent BaseModel class."""
        pass

    def fit(
        self,
        X_train: Union[pd.DataFrame, np.ndarray],
        y_train: Union[pd.Series, list, np.ndarray],
        valid: Tuple[
            Union[pd.DataFrame, np.ndarray], Union[pd.Series, list, np.ndarray]
        ] = None,
    ):

        # Outputs a warning if, at a glance, it doesn't make sense for the dataset to be a classification dataset. The 0.1 can be adjusted.
        if ((np.unique(y_train)).size / (y_train.size)) > 0.1:
            logging.warning(
                " # of unique values / # of total datapoints is above the default threshold of 0.1."
            )

        self.sorted_classes = np.unique(
            y_train
        )  # Sorts the unique values in the array. Ex. [Z,X,Y,Z,X,X,Y] -> [X,Y,Z]
        for classes in self.sorted_classes:
            y_train_class = (y_train == classes).astype(
                int
            )  # Sets the currently iterating class (ex. X) to 1 and all others (ex. Y,Z) to 0, converting the dataset to a binary classification problem.
            self.model = (
                self.classifier.copy()
            )  # Creates a copy of the user-specified model (ex. Random Forest)
            self.model.fit(
                X_train, y_train_class
            )  # Calls the BaseModel fit method, since we now just have a binary classification dataset.
            self.classifiers.append(
                self.model
            )  # Adds the results of the fitted model to an array for later analysis.

    def _predict(self, X: Union[pd.DataFrame, np.ndarray, list, pd.Series]):
        """Empty method to correctly override the predict method of the parent BaseModel class."""
        pass

    def predict(self, X: Union[pd.DataFrame, np.ndarray, list, pd.Series]):
        """
        Args:
            X (Union[pd.DataFrame, np.ndarray, list, pd.Series]): Input data to be predicted (structures + optionally features).

        Returns:
            pd.DataFrame: Predicted values for the provided dataset, with multiple columns for the 2+ different classes and each row representing a different prediction.
        """
        predictions = {}

        # Iterates through two arrays at once in order to build the predictions dictionary
        # The key is the class and the value is the binary classification prediction result for that class vs. all the others.
        for y_class, classifier in zip(self.sorted_classes, self.classifiers):
            predictions[y_class] = classifier.predict(X)

        predictions = pd.DataFrame(predictions)
        predictions = (
            predictions.T / predictions.sum(axis=1)
        ).T  # normalizes outputs between 0 and 1

        return predictions

    def _save(self):
        d = super()._save()
        d.update(
            {
                "classifiers": [saves(classifier) for classifier in self.classifiers],
                "sorted_classes": self.sorted_classes,
            }
        )

    def _load(self, d):
        super()._load(d)
        self.classifiers = [loads(classifier) for classifier in d["classifiers"]]
        self.sorted_classes = d["sorted_classes"]


class BaseSKLearnModel(BaseModel):
    """Base class for creating sklearn-type models, e.g. with a sklearn RandomForestRegressor and RandomForestClassifier.

    Attributes:
        representation (BaseRepresentation): Representation to be used to preprocess the input data.
        regression_model (BaseModel or BaseEstimator): Model to be used for regression tasks.
        classification_model (BaseModel or BaseEstimator): Model to be used for classification tasks.
    """

    @log_arguments
    def __init__(
        self, representation, regression_model, classification_model, log=True, **kwargs
    ):
        self.representation = representation
        self.regression_model = regression_model
        self.classification_model = classification_model
        super().__init__(log=False, **kwargs)

    def preprocess(self, X, y, fit=False):
        if self.representation is None:
            return X
        else:
            return self.representation.convert(X, fit=fit)

    def _fit(self, X_train, y_train):
        values, counts = np.unique(y_train, return_counts=True)
        if len(values) == 2:
            self.setting = "classification"
            self.classification_model.fit(X_train, y_train)
            self.model = self.classification_model
        else:
            self.setting = "regression"
            self.regression_model.fit(X_train, y_train)
            self.model = self.regression_model

    def _predict(self, X):
        return self.model.predict(X)

    def _save(self):
        d = super()._save()
        if self.setting == "classification":
            d.update(
                {"model": self.classification_model._save(), "setting": self.setting}
            )
        else:
            d.update({"model": self.regression_model._save(), "setting": self.setting})
        return d

    def _load(self, d):
        super()._load(d)
        if d["setting"] == "classification":
            self.classification_model._load(d["model"])
            self.model = self.classification_model
        else:
            self.regression_model._load(d["model"])
            self.model = self.regression_model


class BaseErrorModel(BaseClass):
    """Base class for error models.

    Estimates confidence intervals for trained oce models.

    BaseDomainApplicability will be depreciated.

    Methods:
        build: builds the error model from a trained BaseModel and dataset
        calculate: calculates confidence scores from inputs
        fit_valid: fits confidence scores to a trained model and validation dataset
        fit_cv: fits confidence scores to k-fold cross validation on the training dataset
        _fit: fits confidence scores to residuals
        score: returns confidence intervals on a dataset
    """

    def __init__(self):
        pass

    def build(
        self,
        model: BaseModel,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
        y: Union[np.ndarray, list, pd.Series],
    ):
        """Builds the error model with a trained model and training dataset

        Parameters:
            model (BaseModel): trained model
            X (array-like): training features, smiles
            y (array-like): training values
        """
        self.model = model
        self.X_train = X
        self.y_train = np.array(y)
        self.y_pred_train = np.array(self.model.predict(self.X_train)).flatten()

    @abstractmethod
    def calculate(
        self,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
        y_pred: np.ndarray,
    ) -> np.ndarray:
        """To be implemented by the child class; calculates confidence scores from inputs.

        Args:
            X: features, SMILES
            y_pred (1-dimensional np.ndarray): predicted values

        Returns:
            scores (1-dimensional np.ndarray)
        """
        pass

    def fit(
        self,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
        y: Union[np.ndarray, list, pd.Series],
        **kwargs,
    ):
        """Fits confidence scores to an external dataset

        Args:
            X (array-like): features, smiles
            y (array-like): true values
        """
        y_pred = np.array(self.model.predict(X)).flatten()
        residuals = np.abs(np.array(y) - y_pred)

        scores = np.array(self.calculate(X, y_pred))
        self._fit(residuals, scores, **kwargs)

    def fit_cv(
        self,
        n_splits: int = 10,
        test_size_range: Union[tuple, list] = (0.05, 0.5),
        **kwargs,
    ):
        """Fits confidence scores to the training dataset via cross validation.

        Args:
            n_splits (int): number of cross validations
            test_size_range (tuple or list): range of test split proportions
        """
        residuals = None
        scores = None
        splits = np.linspace(*test_size_range, n_splits)

        from sklearn.model_selection import train_test_split

        for test_size in splits:
            print("cross validating with split={}".format(np.round(test_size, 3)))
            X_train, X_test, y_train, y_test = train_test_split(
                self.X_train, self.y_train, test_size=test_size, random_state=0
            )
            model = self.model.copy()
            model.fit(X_train, y_train)
            y_pred_test = np.array(model.predict(X_test)).flatten()
            if residuals is None:
                residuals = np.abs(y_test - y_pred_test)
            else:
                residuals = np.concatenate((residuals, np.abs(y_test - y_pred_test)))

            em = type(self)(*self.args, **self.kwargs)
            em.build(model, X_train, y_train)
            new_scores = em.calculate(X_test, y_pred_test)
            if scores is None:
                scores = new_scores
            else:
                scores = np.concatenate((scores, new_scores))

        self._fit(residuals, scores, **kwargs)

    def _fit(
        self,
        residuals: np.ndarray,
        scores: np.ndarray,
        quantile: float = 0.8,
        method: str = "roll",
        window: int = 100,
        bins: int = 10,
        min_per_bin: int = 5,
        filename: str = "figure.png",
    ):
        """Fits confidence scores to residuals.

        Args:
            residuals (1-dimensional np.ndarray): array of residuals
            scores (1-dimensional np.ndarray): array of confidence scores
            method ({'bin','qbin','roll'}): whether to fit the error model via binning, quantile binning, or rolling quantile
            bins (int): number of bins for binned quantiles
            window (int): window size for rolling quantiles
            quantile (float): confidence interval quantile to capture during fitting
            min_per_bin (int): minimum number of instances per bin
            filename (str): save destination of the fitted plot
        """
        if method == "bin":
            bin_labels = pd.cut(scores, bins, labels=False)
            X = []
            y = []
            for i in range(bins):
                ith_bin = bin_labels == i
                if np.sum(ith_bin) >= min_per_bin:
                    X.append(np.mean(scores[ith_bin]))
                    y.append(pd.Series(residuals[ith_bin]).quantile(quantile))
            X = np.array(X)
            y = np.array(y)
        elif method == "qbin":
            bin_labels = pd.qcut(scores, bins, labels=False, duplicates="drop")
            X = np.array(
                [np.mean(scores[bin_labels == i]) for i in range(max(bin_labels) + 1)]
            )
            y = np.array(
                [
                    pd.Series(residuals[bin_labels == i]).quantile(quantile)
                    for i in range(max(bin_labels) + 1)
                ]
            )
        elif method == "roll":
            results = list(zip(scores, residuals))
            scores, residuals = zip(*sorted(results))
            X = pd.Series(scores).rolling(window).mean()
            y = pd.Series(residuals).rolling(window).quantile(quantile)
            X = np.array(X[~np.isnan(X)])
            y = np.array(y[~np.isnan(y)])
        else:
            raise NameError("method {} is not recognized".format(method))

        import matplotlib.pyplot as plt
        from scipy.optimize import curve_fit

        funcs = [
            lambda x, a, b: a * np.exp(b * x),
            lambda x, a, b: a * np.log(x) + b,
            lambda x, a, b: a * x**b,
            lambda x, a, b: a * x + b,
        ]
        min_mse = None
        for func in funcs:
            try:
                popt = curve_fit(func, X, y)[0]
            except RuntimeError:
                continue
            y_pred = func(X, *popt)
            try:
                mse = mean_squared_error(y, y_pred)
            except ValueError:
                continue
            if min_mse is None or mse < min_mse:
                opt_func = func
                opt_popt = popt
                min_mse = mse
        X_min, X_max = np.min(X), np.max(X)
        self.reg = np.vectorize(
            lambda x: opt_func(max(min(x, X_max), X_min), *opt_popt)
        )

        plt.xlabel(self.__class__.__name__)
        plt.ylabel("Absolute Error")
        sorted_scores = np.sort(scores)

        plt.scatter(scores, residuals, c="black", alpha=0.2)
        plt.scatter(X, y, c="blue")
        plt.plot(sorted_scores, self.reg(sorted_scores), c="red")
        plt.savefig(filename)

        self.residuals = residuals
        self.scores = scores
        self.method = method
        self.bins = bins
        self.window = window
        self.quantile = quantile
        self.min_per_bin = min_per_bin
        self.filename = filename

    def score(
        self,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
    ) -> np.ndarray:
        """Calculates confidence scores on a dataset.

        Args:
            X (array-like): dataset being evaluated, sequence of SMILES

        Returns:
            a list of confidence intervals for each input
        """
        assert hasattr(self, "reg"), "error model not yet fitted"

        y_pred = np.array(self.model.predict(X)).flatten()
        scores = self.calculate(X, y_pred)

        return self.reg(scores)

    def copy(self) -> BaseErrorModel:
        """returns a copy of itself

        Returns:
            BaseErrorModel: copied instance of itself
        """
        if hasattr(self, "args"):
            return type(self)(*self.args, **self.kwargs)
        else:
            return type(self)()

    def _save(self) -> dict:
        d = {}
        if hasattr(self, "model") and not self is getattr(
            self.model, "error_model", None
        ):
            d.update({"model": saves(self.model)})
        if hasattr(self, "X_train"):
            d.update({"X_train": self.X_train})
            d.update({"y_train": self.y_train})
            d.update({"y_pred_train": self.y_pred_train})
        if hasattr(self, "residuals"):
            d.update({"residuals": self.residuals})
            d.update({"scores": self.scores})
            d.update({"method": self.method})
            d.update({"bins": self.bins})
            d.update({"window": self.window})
            d.update({"quantile": self.quantile})
            d.update({"min_per_bin": self.min_per_bin})
            d.update({"filename": self.filename})
        return d

    def _load(self, d) -> None:
        if "model" in d.keys():
            self.model = loads(d["model"])
        if "X_train" in d.keys():
            self.X_train = d["X_train"]
            self.y_train = d["y_train"]
            self.y_pred_train = d["y_pred_train"]
        if "residuals" in d.keys():
            residuals = d["residuals"]
            scores = d["scores"]
            method = d["method"]
            bins = d["bins"]
            window = d["window"]
            quantile = d["quantile"]
            min_per_bin = d["min_per_bin"]
            filename = d["filename"]

            self._fit(
                residuals,
                scores,
                method=method,
                bins=bins,
                window=window,
                quantile=quantile,
                min_per_bin=min_per_bin,
                filename=filename,
            )


class BaseAggregateErrorModel(BaseErrorModel):
    """Base class for aggregate error models.

    Parameters:
        error_models (BaseErrorModel or list of BaseErrorModel): list of error models to be aggregated
    """

    @log_arguments
    def __init__(
        self, error_models: Union[BaseErrorModel, List[BaseErrorModel]], **kwargs
    ):
        if not isinstance(error_models, list):
            error_models = [error_models]
        self.error_models = error_models
        self.kwargs = kwargs

    def build(
        self,
        model: BaseModel,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
        y: Union[np.ndarray, list, pd.Series],
    ):
        for error_model in self.error_models:
            error_model.build(model, X, y)
        super().build(model, X, y)

    def train(
        self,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
        y: Union[np.ndarray, list, pd.Series],
    ):
        """Trains the aggregate model to a validation dataset.

        Args:
            X (array-like): features, smiles
            y (array-like): true values
        """
        y_pred = np.array(self.model.predict(X)).flatten()
        residuals = np.abs(np.array(y) - y_pred)

        scores = np.transpose(
            np.stack([self.get_scores(em, X, y_pred) for em in self.error_models])
        )
        self._train(scores, residuals)

    def train_cv(
        self, n_splits: int = 9, test_size_range: Union[tuple, list] = (0.1, 0.9)
    ):
        """Trains the aggregate model to the training dataset via cross validation.

        Args:
            n_splits (int): number of cross validations
            test_size_range (tuple or list): range of test split proportions
        """
        residuals = None
        scores = None
        splits = np.linspace(*test_size_range, n_splits)

        from sklearn.model_selection import train_test_split

        for test_size in splits:
            print("cross validating with split={}".format(np.round(test_size, 3)))
            X_train, X_test, y_train, y_test = train_test_split(
                self.X_train, self.y_train, test_size=test_size, random_state=0
            )
            model = self.model.copy()
            model.fit(X_train, y_train)
            y_pred_test = np.array(model.predict(X_test)).flatten()
            if residuals is None:
                residuals = np.abs(y_test - y_pred_test)
            else:
                residuals = np.concatenate((residuals, np.abs(y_test - y_pred_test)))

            new_scores = []
            for error_model in self.error_models:
                em = type(error_model)(*error_model.args, **error_model.kwargs)
                em.build(model, X_train, y_train)
                new_scores.append(self.get_scores(em, X_test, y_pred_test))
            new_scores = np.transpose(np.stack(new_scores))
            if scores is None:
                scores = new_scores
            else:
                scores = np.concatenate((scores, new_scores))

        self._train(scores, residuals)

    def calculate(
        self,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
        y_pred: np.ndarray,
        **kwargs,
    ) -> np.ndarray:
        """Computes aggregate error model score from inputs.

        Args:
            X: features, smiles
            y_pred: predicted values
        """
        assert hasattr(self, "aggregate_model"), "aggregate model not yet trained"

        scores = [self.get_scores(em, X, y_pred) for em in self.error_models]
        scores = np.transpose(np.stack(scores))
        return self._predict(scores, **kwargs)

    def get_scores(
        self,
        em: BaseErrorModel,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
        y_pred: np.ndarray,
    ) -> np.ndarray:
        """Gets confidence scores from a model, inputs, and keyword arguments.

        Args:
            em (BaseErrorModel): error model
            X: features, smiles
            y_pred: predicted values
        """
        return em.calculate(X, y_pred)

    @abstractmethod
    def _train(self, scores: np.ndarray, residuals: np.ndarray, **kwargs):
        """To be implemented by the child class; trains the aggregate model based on residuals
        and confidence scores and assigns the trained model to `self.aggregate_model`.

        Args:
            scores (np.ndarray): confidence scores
            residuals (np.ndarray): residuals
        """
        pass

    @abstractmethod
    def _predict(self, scores: np.ndarray, **kwargs) -> np.ndarray:
        """To be implemented by the child class; predicts from the aggregate model based on
        confdience scores.

        Args:
            scores (np.ndarray): confidence scores
        """
        pass

    def _save(self) -> dict:
        d = super()._save()
        if hasattr(self, "aggregate_model"):
            d.update({"aggregate_model": saves(self.aggregate_model)})
        return d

    def _load(self, d) -> None:
        super()._load(d)
        if "aggregate_model" in d.keys():
            self.aggregate_model = loads(d["aggregate_model"])



"""
Code taken in part from https://github.com/DSPsleeporg/smiles-transformer under
the following license:

MIT License

Copyright (c) 2019 Shion Honda

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
Footer
"""

import argparse
import pickle
from collections import Counter

import olorenchemengine as oce
from olorenchemengine.internal import mock_imports


class TorchVocab(object):
    """
    :property freqs: collections.Counter, コーパス中の単語の出現頻度を保持するオブジェクト
    :property stoi: collections.defaultdict, string → id の対応を示す辞書
    :property itos: collections.defaultdict, id → string の対応を示す辞書
    """

    def __init__(
        self,
        counter,
        max_size=None,
        min_freq=1,
        specials=["<pad>", "<oov>"],
        vectors=None,
        unk_init=None,
        vectors_cache=None,
    ):
        """
        :param counter: collections.Counter, データ中に含まれる単語の頻度を計測するためのcounter
        :param max_size: int, vocabularyの最大のサイズ. Noneの場合は最大値なし. defaultはNone
        :param min_freq: int, vocabulary中の単語の最低出現頻度. この数以下の出現回数の単語はvocabularyに加えられない.
        :param specials: list of str, vocabularyにあらかじめ登録するtoken
        :param vectors: list of vectors, 事前学習済みのベクトル. ex)Vocab.load_vectors
        """
        self.freqs = counter
        counter = counter.copy()
        min_freq = max(min_freq, 1)

        self.itos = list(specials)
        # special tokensの出現頻度はvocabulary作成の際にカウントされない
        for tok in specials:
            del counter[tok]

        max_size = None if max_size is None else max_size + len(self.itos)

        # まず頻度でソートし、次に文字順で並び替える
        words_and_frequencies = sorted(counter.items(), key=lambda tup: tup[0])
        words_and_frequencies.sort(key=lambda tup: tup[1], reverse=True)

        # 出現頻度がmin_freq未満のものはvocabに加えない
        for word, freq in words_and_frequencies:
            if freq < min_freq or len(self.itos) == max_size:
                break
            self.itos.append(word)

        # dictのk,vをいれかえてstoiを作成する
        self.stoi = {tok: i for i, tok in enumerate(self.itos)}

        self.vectors = None
        if vectors is not None:
            self.load_vectors(vectors, unk_init=unk_init, cache=vectors_cache)
        else:
            assert unk_init is None and vectors_cache is None

    def __eq__(self, other):
        if self.freqs != other.freqs:
            return False
        if self.stoi != other.stoi:
            return False
        if self.itos != other.itos:
            return False
        if self.vectors != other.vectors:
            return False
        return True

    def __len__(self):
        return len(self.itos)

    def vocab_rerank(self):
        self.stoi = {word: i for i, word in enumerate(self.itos)}

    def extend(self, v, sort=False):
        words = sorted(v.itos) if sort else v.itos
        for w in words:
            if w not in self.stoi:
                self.itos.append(w)
                self.stoi[w] = len(self.itos) - 1


class Vocab(TorchVocab):
    def __init__(self, counter, max_size=None, min_freq=1):
        self.pad_index = 0
        self.unk_index = 1
        self.eos_index = 2
        self.sos_index = 3
        self.mask_index = 4
        super().__init__(
            counter,
            specials=["<pad>", "<unk>", "<eos>", "<sos>", "<mask>"],
            max_size=max_size,
            min_freq=min_freq,
        )

    # override用
    def to_seq(self, sentece, seq_len, with_eos=False, with_sos=False) -> list:
        pass

    # override用
    def from_seq(self, seq, join=False, with_pad=False):
        pass

    @staticmethod
    def load_vocab(vocab_path: str) -> "Vocab":
        with open(vocab_path, "rb") as f:
            return pickle.load(f)

    def save_vocab(self, vocab_path):
        with open(vocab_path, "wb") as f:
            pickle.dump(self, f)


# テキストファイルからvocabを作成する
class WordVocab(Vocab):
    def __init__(self, texts, max_size=None, min_freq=1):
        print("Building Vocab")
        counter = Counter()
        for line in texts:
            if isinstance(line, list):
                words = line
            else:
                words = line.replace("\n", "").replace("\t", "").split()

            for word in words:
                counter[word] += 1
        super().__init__(counter, max_size=max_size, min_freq=min_freq)

    def to_seq(
        self, sentence, seq_len=None, with_eos=False, with_sos=False, with_len=False
    ):
        if isinstance(sentence, str):
            sentence = sentence.split()

        seq = [self.stoi.get(word, self.unk_index) for word in sentence]

        if with_eos:
            seq += [self.eos_index]  # this would be index 1
        if with_sos:
            seq = [self.sos_index] + seq

        origin_seq_len = len(seq)

        if seq_len is None:
            pass
        elif len(seq) <= seq_len:
            seq += [self.pad_index for _ in range(seq_len - len(seq))]
        else:
            seq = seq[:seq_len]

        return (seq, origin_seq_len) if with_len else seq

    def from_seq(self, seq, join=False, with_pad=False):
        words = [
            self.itos[idx] if idx < len(self.itos) else "<%d>" % idx
            for idx in seq
            if not with_pad or idx != self.pad_index
        ]

        return " ".join(words) if join else words

    @staticmethod
    def load_vocab(vocab_path: str):
        with open(vocab_path, "rb") as f:
            return pickle.load(f)


import math

import numpy as np
import pandas as pd
import torch
from torch import nn, optim
from torch.autograd import Variable
from torch.nn import functional as F
from torch.utils.data import DataLoader
from tqdm import tqdm

PAD = 0
UNK = 1
EOS = 2
SOS = 3
MASK = 4


class PositionalEncoding(nn.Module):
    "Implement the PE function. No batch support?"

    def __init__(self, d_model, dropout, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        # Compute the positional encodings once in log space.
        pe = torch.zeros(max_len, d_model)  # (T,H)
        position = torch.arange(0.0, max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0.0, d_model, 2) * -(math.log(10000.0) / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.register_buffer("pe", pe)

    def forward(self, x):
        x = x + Variable(self.pe[:, : x.size(1)], requires_grad=False)
        return self.dropout(x)


class TrfmSeq2seq(nn.Module):
    def __init__(self, in_size, hidden_size, out_size, n_layers, dropout=0.1):
        super(TrfmSeq2seq, self).__init__()
        self.in_size = in_size
        self.hidden_size = hidden_size
        self.embed = nn.Embedding(in_size, hidden_size)
        self.embed.to(oce.CONFIG["DEVICE"])

        self.pe = PositionalEncoding(hidden_size, dropout)
        self.pe.to(oce.CONFIG["DEVICE"])

        self.trfm = nn.Transformer(
            d_model=hidden_size,
            nhead=4,
            num_encoder_layers=n_layers,
            num_decoder_layers=n_layers,
            dim_feedforward=hidden_size,
        )
        self.trfm.to(oce.CONFIG["DEVICE"])

        self.out = nn.Linear(hidden_size, out_size)
        self.trfm.to(oce.CONFIG["DEVICE"])

    def forward(self, src):
        # src: (T,B)
        embedded = self.embed(src)  # (T,B,H)
        embedded = self.pe(embedded)  # (T,B,H)
        hidden = self.trfm(embedded, embedded)  # (T,B,H)
        out = self.out(hidden)  # (T,B,V)
        out = F.log_softmax(out, dim=2)  # (T,B,V)
        return out  # (T,B,V)

    def _encode(self, src):
        # src: (T,B)
        embedded = self.embed(src)  # (T,B,H)
        embedded = self.pe(embedded)  # (T,B,H)
        output = embedded
        for i in range(self.trfm.encoder.num_layers - 1):
            output = self.trfm.encoder.layers[i](output, None)  # (T,B,H)
        penul = output.detach().cpu().numpy()
        output = self.trfm.encoder.layers[-1](output, None)  # (T,B,H)
        if self.trfm.encoder.norm:
            output = self.trfm.encoder.norm(output)  # (T,B,H)
        output = output.detach().cpu().numpy()
        # mean, max, first*2
        return np.hstack(
            [
                np.mean(output, axis=0),
                np.max(output, axis=0),
                output[0, :, :],
                penul[0, :, :],
            ]
        )  # (B,4H)

    def encode(self, src):
        # src: (T,B)
        batch_size = src.shape[1]
        if batch_size <= 100:
            return self._encode(src)
        else:  # Batch is too large to load
            print(
                "There are {:d} molecules. It will take a little time.".format(
                    batch_size
                )
            )
            st, ed = 0, 100
            out = self._encode(src[:, st:ed])  # (B,4H)
            while ed < batch_size:
                st += 100
                ed += 100
                out = np.concatenate([out, self._encode(src[:, st:ed])], axis=0)
            return out


from torch.utils.data import Dataset


def split(sm):
    """
    function: Split SMILES into words. Care for Cl, Br, Si, Se, Na etc.
    input: A SMILES
    output: A string with space between words
    """
    arr = []
    i = 0
    while i < len(sm) - 1:
        if not sm[i] in [
            "%",
            "C",
            "B",
            "S",
            "N",
            "R",
            "X",
            "L",
            "A",
            "M",
            "T",
            "Z",
            "s",
            "t",
            "H",
            "+",
            "-",
            "K",
            "F",
        ]:
            arr.append(sm[i])
            i += 1
        elif sm[i] == "%":
            arr.append(sm[i : i + 3])
            i += 3
        elif sm[i] == "C" and sm[i + 1] == "l":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "C" and sm[i + 1] == "a":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "C" and sm[i + 1] == "u":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "B" and sm[i + 1] == "r":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "B" and sm[i + 1] == "e":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "B" and sm[i + 1] == "a":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "B" and sm[i + 1] == "i":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "S" and sm[i + 1] == "i":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "S" and sm[i + 1] == "e":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "S" and sm[i + 1] == "r":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "N" and sm[i + 1] == "a":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "N" and sm[i + 1] == "i":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "R" and sm[i + 1] == "b":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "R" and sm[i + 1] == "a":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "X" and sm[i + 1] == "e":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "L" and sm[i + 1] == "i":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "A" and sm[i + 1] == "l":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "A" and sm[i + 1] == "s":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "A" and sm[i + 1] == "g":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "A" and sm[i + 1] == "u":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "M" and sm[i + 1] == "g":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "M" and sm[i + 1] == "n":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "T" and sm[i + 1] == "e":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "Z" and sm[i + 1] == "n":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "s" and sm[i + 1] == "i":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "s" and sm[i + 1] == "e":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "t" and sm[i + 1] == "e":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "H" and sm[i + 1] == "e":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "+" and sm[i + 1] == "2":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "+" and sm[i + 1] == "3":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "+" and sm[i + 1] == "4":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "-" and sm[i + 1] == "2":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "-" and sm[i + 1] == "3":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "-" and sm[i + 1] == "4":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "K" and sm[i + 1] == "r":
            arr.append(sm[i : i + 2])
            i += 2
        elif sm[i] == "F" and sm[i + 1] == "e":
            arr.append(sm[i : i + 2])
            i += 2
        else:
            arr.append(sm[i])
            i += 1
    if i == len(sm) - 1:
        arr.append(sm[i])
    return " ".join(arr)


class Seq2seqDataset(Dataset):
    def __init__(self, smiles, vocab, seq_len=220):
        self.smiles = smiles
        self.vocab = vocab
        self.seq_len = seq_len

    def __len__(self):
        return len(self.smiles)

    def __getitem__(self, item):
        sm = self.smiles[item]
        sm = split(sm).split()
        content = [self.vocab.stoi.get(token, self.vocab.unk_index) for token in sm]
        X = [self.vocab.sos_index] + content + [self.vocab.eos_index]
        padding = [self.vocab.pad_index] * (self.seq_len - len(X))
        X.extend(padding)
        return torch.tensor(X)



"""
Wraps the model presented in `Analyzing Learned Molecular Representations for Property Prediction <hhttps://doi.org/10.1021/acs.jcim.9b00237>`.
Here, we adapt its PyTorch Geometric implementation as in the `Github repository <https://github.com/itakigawa/pyg_chemprop>`

"""

import io

import numpy as np
import torch
import torch.nn as nn
import torch.utils.data
from rdkit import Chem
from torch_geometric.data import Data, Dataset
from torch_geometric.data.data import size_repr
from torch_geometric.loader import DataLoader
from torch_geometric.nn import global_mean_pool
from torch_scatter import scatter_sum
from tqdm import tqdm

from olorenchemengine.base_class import BaseModel, QuantileTransformer, log_arguments
from olorenchemengine.internal import mock_imports
from olorenchemengine.representations import (
    AtomFeaturizer,
    BondFeaturizer,
    TorchGeometricGraph,
)


class RevIndexedData(Data):
    def __init__(self, orig):
        super(RevIndexedData, self).__init__()
        if orig:
            for key in orig.keys:
                self[key] = orig[key]
            edge_index = self["edge_index"]
            revedge_index = torch.zeros(edge_index.shape[1]).long()
            for k, (i, j) in enumerate(zip(*edge_index)):
                edge_to_i = edge_index[1] == i
                edge_from_j = edge_index[0] == j
                revedge_index[k] = torch.where(edge_to_i & edge_from_j)[0].item()
            self["revedge_index"] = revedge_index

    def __inc__(self, key, value, *args, **kwargs):
        if key == "revedge_index":
            if len(self.revedge_index) > 0:
                return self.revedge_index.max().item() + 1
            else:
                return 0
        else:
            return super().__inc__(key, value)

    def __repr__(self):
        cls = str(self.__class__.__name__)
        has_dict = any([isinstance(item, dict) for _, item in self])

        if not has_dict:
            info = [size_repr(key, item) for key, item in self]
            return "{}({})".format(cls, ", ".join(info))
        else:
            info = [size_repr(key, item, indent=2) for key, item in self]
            return "{}(\n{}\n)".format(cls, ",\n".join(info))


class RevIndexedDataset(Dataset):
    def __init__(self, orig):
        super(RevIndexedDataset, self).__init__()
        self.dataset = [RevIndexedData(data) for data in tqdm(orig)]

    def __getitem__(self, idx):
        return self.dataset[idx]

    def __len__(self):
        return len(self.dataset)


def onek_encoding_unk(value, choices):
    encoding = [0] * (len(choices) + 1)
    index = choices.index(value) if value in choices else -1
    encoding[index] = 1
    return encoding


class ChemProp_AF(AtomFeaturizer):
    @property
    def length(self):
        return 151

    def convert(self, atom: Chem.Atom) -> np.ndarray:

        ATOM_FEATURES = {
            "atomic_num": list(range(1, 119)),
            "degree": [0, 1, 2, 3, 4, 5],
            "formal_charge": [-1, -2, 1, 2, 0],
            "chiral_tag": [0, 1, 2, 3],
            "num_Hs": [0, 1, 2, 3, 4],
            "hybridization": [
                Chem.rdchem.HybridizationType.SP,
                Chem.rdchem.HybridizationType.SP2,
                Chem.rdchem.HybridizationType.SP3,
                Chem.rdchem.HybridizationType.SP3D,
                Chem.rdchem.HybridizationType.SP3D2,
            ],
        }
        features = (
            onek_encoding_unk(atom.GetAtomicNum() - 1, ATOM_FEATURES["atomic_num"])
            + onek_encoding_unk(atom.GetTotalDegree(), ATOM_FEATURES["degree"])
            + onek_encoding_unk(atom.GetFormalCharge(), ATOM_FEATURES["formal_charge"])
            + onek_encoding_unk(int(atom.GetChiralTag()), ATOM_FEATURES["chiral_tag"])
            + onek_encoding_unk(int(atom.GetTotalNumHs()), ATOM_FEATURES["num_Hs"])
            + onek_encoding_unk(
                int(atom.GetHybridization()), ATOM_FEATURES["hybridization"]
            )
            + [1 if atom.GetIsAromatic() else 0]
            + [atom.GetMass() * 0.01]
        )  # scaled to about the same range as other features
        return np.array(features)


class ChemProp_BF(BondFeaturizer):
    @property
    def length(self):
        return 14

    def convert(self, bond: Chem.Bond) -> np.ndarray:
        if bond is None:
            fbond = [1] + [0] * (13)
        else:
            bt = bond.GetBondType()
            fbond = [
                0,  # bond is not None
                bt == Chem.rdchem.BondType.SINGLE,
                bt == Chem.rdchem.BondType.DOUBLE,
                bt == Chem.rdchem.BondType.TRIPLE,
                bt == Chem.rdchem.BondType.AROMATIC,
                (bond.GetIsConjugated() if bt is not None else 0),
                (bond.IsInRing() if bt is not None else 0),
            ]
            fbond += onek_encoding_unk(int(bond.GetStereo()), list(range(6)))
        return np.array(fbond)


class ChemPropDataLoading(TorchGeometricGraph):
    @log_arguments
    def __init__(self):
        super().__init__(ChemProp_AF(), ChemProp_BF(), log=False)

    def _convert(self, smiles, y=None, **kwargs):
        data = super()._convert(smiles, y=y, **kwargs)
        return RevIndexedData(data)


def initialize_weights(model: nn.Module) -> None:
    """
    Initializes the weights of a model in place.
    :param model: An PyTorch model.
    """
    for param in model.parameters():
        if param.dim() == 1:
            nn.init.constant_(param, 0)
        else:
            nn.init.xavier_normal_(param)


def directed_mp(message, edge_index, revedge_index):
    m = scatter_sum(message, edge_index[1], dim=0)
    m_all = m[edge_index[0]]
    m_rev = message[revedge_index]
    return m_all - m_rev


def aggregate_at_nodes(num_nodes, message, edge_index):
    m = scatter_sum(message, edge_index[1], dim=0, dim_size=num_nodes)
    return m[torch.arange(num_nodes)]


#### Key Message-Passing algorithm in the model ####
class DMPNNEncoder(nn.Module):
    def __init__(self, hidden_size, node_fdim, edge_fdim, depth=3):
        super(DMPNNEncoder, self).__init__()
        self.act_func = nn.ReLU()
        self.W1 = nn.Linear(node_fdim + edge_fdim, hidden_size, bias=False)
        self.W2 = nn.Linear(hidden_size, hidden_size, bias=False)
        self.W3 = nn.Linear(node_fdim + hidden_size, hidden_size, bias=True)
        self.depth = depth

    def forward(self, data):
        x, edge_index, revedge_index, edge_attr, num_nodes, batch = (
            data.x,
            data.edge_index,
            data.revedge_index,
            data.edge_attr,
            data.num_nodes,
            data.batch,
        )

        # initialize messages on edges
        init_msg = torch.cat([x[edge_index[0]], edge_attr], dim=1).float()
        h0 = self.act_func(self.W1(init_msg))

        # directed message passing over edges
        h = h0
        for _ in range(self.depth - 1):
            m = directed_mp(h, edge_index, revedge_index)
            h = self.act_func(h0 + self.W2(m))

        # aggregate in-edge messages at nodes
        v_msg = aggregate_at_nodes(num_nodes, h, edge_index)

        z = torch.cat([x, v_msg], dim=1)
        node_attr = self.act_func(self.W3(z))

        # readout: pyg global pooling
        return global_mean_pool(node_attr, batch)


#### General ChemProp model training method, to be used with _fit in the ChemPropModel() class implementation. ####
def train(config, loader, setting, device=torch.device("cpu")):
    criterion = config["loss"]
    model = config["model"]
    optimizer = config["optimizer"]
    scheduler = config["scheduler"]

    if setting == "classification":
        criterion = nn.BCEWithLogitsLoss(reduction="none")
    else:
        criterion = nn.MSELoss()

    model = model.to(device)
    model.train()
    for batch in loader:
        batch = batch.to(device)

        optimizer.zero_grad()
        out = model(batch)
        loss = criterion(torch.squeeze(out, 1), batch.y.float())
        loss.sum().backward()

        optimizer.step()
        scheduler.step()


#### General ChemProp model prediction method, to be used with _predict in the ChemPropModel() class implementation. ####
def predict(config, loader, setting="classification", device=torch.device("cpu")):
    model = config["model"]

    model = model.to(device)
    model.eval()
    y_pred = []
    for batch in tqdm(loader, total=len(loader)):
        batch = batch.to(device)
        with torch.no_grad():
            if setting == "classification":
                batch_preds = torch.sigmoid(model(batch))
            else:
                batch_preds = model(batch)
        y_pred.extend(batch_preds)

    predictions = np.concatenate([y.cpu().numpy() for y in y_pred])
    return predictions.flatten()


class ChemPropModel(BaseModel):
    """ChemProp is the model presented in `Analyzing Learned Molecular Representations for Property Prediction <hhttps://doi.org/10.1021/acs.jcim.9b00237>` _.
    Here, we adapt its PyTorch Geometric implementation as in the `Github repository <https://github.com/itakigawa/pyg_chemprop>`_

    "ChemProp" is a simple but effective Graph Neural Network (GNN) for Molecular Property Prediction. The PyG implementation makes it compatible with Oloren AI software.

    Attributes:
        representation (TorchGeometricGraph): Returns smiles-inputted data as PyTorch graph objects for use in the ChemProp model
        setting (str): whether the model is a "classification" model or a "regression" model
        optimizer: function to modify model parameters such as weights and learning rate; we use Adam.
        criterion: loss function; we use BCEWithLogitsLoss for classification and MSELoss for regression.
        scheduler: reduces LR as number of training epochs increases; we use PyTorch's OneCycleLR.

    Parameters:
        dropout_rate (float): fraction of layer outputs that are randomly ignored; default = 0.
        epochs (int): number of complete passes of the training dataset through the model; default = 3.
        batch_size (int): number of training examples utilized in one model iteration; default = 50.
        lr (float): amount that the weights are updated during training; default = 1e-3.
        hidden_size (int): number of hidden neurons between input and output; default = 300.
        depth (int): number of layers between input and output; default = 3.
    """

    @log_arguments
    def __init__(
        self,
        dropout_rate: float = 0.0,
        epochs: int = 3,
        batch_size: int = 50,
        lr: float = 1e-3,
        hidden_size: int = 300,
        depth: int = 3,
        map_location="cuda:0",
        **kwargs
    ):

        self.representation = ChemPropDataLoading()

        self.epochs = epochs
        self.batch_size = batch_size
        self.lr = lr
        self.hidden_size = hidden_size
        self.dropout_rate = dropout_rate

        self.map_location = map_location
        self.device = torch.device(self.map_location)

        AF = ChemProp_AF()
        BF = ChemProp_BF()

        self.model = nn.Sequential(
            DMPNNEncoder(self.hidden_size, AF.length, BF.length, depth),
            nn.Sequential(
                nn.Dropout(p=self.dropout_rate, inplace=False),
                nn.Linear(hidden_size, hidden_size, bias=True),
                nn.ReLU(),
                nn.Dropout(p=self.dropout_rate, inplace=False),
                nn.Linear(hidden_size, 1, bias=True),
            ),
        )
        initialize_weights(self.model)

        super().__init__(preprocessor=QuantileTransformer(), **kwargs)
        self.model.to(self.device)

        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=1e-4)
        self.criterion = nn.BCEWithLogitsLoss()

    def preprocess(self, X, y, **kwargs):
        if y is None:
            y = [None] * len(X)
        return self.representation.convert(X, ys=y)

    def _fit(self, X, y, **kwargs):

        loader = DataLoader(X, batch_size=self.batch_size, shuffle=True)

        self.scheduler = torch.optim.lr_scheduler.OneCycleLR(
            self.optimizer, max_lr=self.lr, steps_per_epoch=len(X), epochs=self.epochs
        )
        self.config = {
            "loss": self.criterion,
            "model": self.model,
            "optimizer": self.optimizer,
            "scheduler": self.scheduler,
        }

        if len(np.unique(y, return_counts=True)) > 2 and (
            ((np.unique(y)).size / (y.size)) > 0.1
        ):
            self.setting = "regression"

        for epoch in range(self.epochs):
            train(self.config, loader, self.setting)

    def _predict(self, X, **kwargs):
        loader = DataLoader(X, batch_size=self.batch_size, shuffle=False)

        self.scheduler = torch.optim.lr_scheduler.OneCycleLR(
            self.optimizer, max_lr=self.lr, steps_per_epoch=len(X), epochs=self.epochs
        )
        self.config = {
            "loss": self.criterion,
            "model": self.model,
            "optimizer": self.optimizer,
            "scheduler": self.scheduler,
        }
        return predict(self.config, loader, self.setting)

    def _save(self) -> str:
        d = super()._save()
        buffer = io.BytesIO()
        torch.save(self.model, buffer)
        d.update({"save": buffer.getvalue()})
        return d

    def _load(self, d):
        super()._load(d)
        self.model = torch.load(io.BytesIO(d["save"]))



from timeit import default_timer as timer
from typing import List

from tqdm import tqdm

import olorenchemengine as oce

from .base_class import *
from .dataset import *


def TOP_MODELS_ADMET() -> List[BaseModel]:
    """
    Returns a list of the top models from the ADMET dataset.

    Returns:
        List[BaseModel]: A list of the top models from the ADMET dataset.
    """
    df = pd.read_csv(download_public_file("ModelDatabase/ModelDatabase_small.csv"))

    return [oce.create_BC(mp) for mp in df["Parameters"]]


class BaseModelManager(BaseClass):
    """BaseModelManager is the base class for all model managers.

    Parameters:
        dataset (Dataset): The dataset to use for training and testing.
        metrics (List[str]): A list of metrics to use.
        verbose (bool): Whether or not to print progress.
        file_path (str): The path to the model_database.
        log (bool): Whether or not to log to the model_database.
    """

    @log_arguments
    def __init__(
        self,
        dataset: BaseDataset,
        metrics: List[str],
        file_path: str = None,
        primary_metric: str = None,
        verbose=True,
        log=True,
    ):
        self.dataset = dataset

        if not isinstance(metrics, list):
            metrics = [metrics]
        assert sum([m in metric_functions.keys() for m in metrics]) == len(
            metrics
        ), "metrics must be either defined classification or regression metrics"
        self.metrics = metrics

        # Parameters for saving and storing the best model
        if primary_metric is None:
            self.primary_metric = metrics[0]
        else:
            self.primary_metric = primary_metric
        self.best_model = None
        self.best_primary_metric = float("inf")

        self.model_database = pd.DataFrame(
            columns=["Model Name", "Model Parameters", "Fitting Time"] + self.metrics
        )

        self.file_path = file_path
        if not self.file_path is None and os.path.exists(self.file_path):
            with open(self.file_path, "rb") as f:
                d = pickle.load(f)
                self._load(d["instance_save"])

        self.verbose = verbose

    def get_dataset(self):
        return self.dataset

    def primary_metric(self):
        return self.primary_metric

    def run(
        self, models: Union[BaseModel, List[BaseModel]], return_models: bool = False
    ):
        """Runs the model on the dataset and saves the results to the model_database.

        Parameters:
            models (Union[BaseModel,List[BaseModel]]): The model(s) to run.
            return_models (bool): Whether or not to return the trained models.
        """
        import joblib

        if not isinstance(models, list):
            models = [models]

        model_outputs = []
        primary_metric_outputs = []
        print(models)
        for model in tqdm(models) if self.verbose else models:
            model = model.copy()
            model_parameters = parameterize(model)
            model_name = model_name_from_params(model_parameters)

            print("Fitting model")
            start = timer()
            model.fit(*self.dataset.train_dataset)
            end = timer()
            print("Evaluating model")

            if len(self.dataset.valid_dataset[0]) > 0:
                eval_set = self.dataset.valid_dataset
            else:
                eval_set = self.dataset.test_dataset

            y_pred = model.predict(eval_set[0])
            y = eval_set[1]

            stats = {
                metric: metric_functions[metric](y, y_pred) for metric in self.metrics
            }

            if not self.file_path is None and os.path.exists(self.file_path):
                with open(self.file_path, "rb") as f:
                    d = pickle.load(f)
                    self._load(d["instance_save"])

            self.model_database = self.model_database.append(
                {
                    "Model Name": model_name,
                    "Model Parameters": str(model_parameters),
                    "Fitting Time": end - start,
                    **stats,
                },
                ignore_index=True,
            )

            train_dataset = self.dataset.train_dataset
            dataset_hash = joblib.hash(train_dataset[0]) + joblib.hash(train_dataset[1])

            # LOGGING START

            import requests

            requests.get(
                f"https://api.oloren.ai/firestore/log_model_performance",
                params={
                    "name": model.name,
                    "dataset_hash": dataset_hash,
                    "metrics": json.dumps(stats),
                    "metric_direction": json.dumps(metric_direction),
                    "params": json.dumps(parameterize(model)),
                },
            )

            # LOGGING END

            # Using metric direction,
            if metric_direction[self.primary_metric] == "higher":
                primary_metric_outputs.append(
                    -1 * metric_functions[self.primary_metric](y, y_pred)
                )
            else:
                primary_metric_outputs.append(
                    metric_functions[self.primary_metric](y, y_pred)
                )
            if primary_metric_outputs[-1] < self.best_primary_metric:
                self.best_model = model
                self.best_primary_metric = primary_metric_outputs[-1]

            if not self.file_path is None:
                with open(self.file_path, "w+") as f:
                    oce.save(self, self.file_path)

            if return_models:
                model_outputs.append(model)

        if return_models:
            if len(model_outputs) == 1:
                return primary_metric_outputs[0], model_outputs[0]
            else:
                return primary_metric_outputs, model_outputs
        else:
            if len(primary_metric_outputs) == 1:
                return primary_metric_outputs[0]
            else:
                return primary_metric_outputs

    def get_model_database(self):

        return self.model_database


class ModelManager(BaseModelManager):
    """ModelManager is the class that tracks model development against a specified
    dataset. It is responsible for saving parameter settings and metrics.

    Parameters:
        dataset (BaseDataset): The dataset to use for model development.
        metrics (list[Str]): The metrics to track e.g. ROC AUC, Root Mean Squared Error.
        autosave (str): The path to save the model manager to. Optional.
    """

    @log_arguments
    def __init__(
        self,
        dataset: BaseDataset,
        metrics: List[str],
        file_path: str = None,
        primary_metric: str = None,
        verbose=True,
        log=True,
    ):
        self.name = oce.model_name_from_model(self)
        super().__init__(
            dataset,
            metrics,
            file_path=file_path,
            primary_metric=primary_metric,
            verbose=verbose,
            log=False,
        )

    def _save(self) -> dict:
        return {
            "model_database": self.model_database.to_dict(orient="records"),
            "best_model_params": parameterize(self.best_model),
            "best_model": self.best_model._save(),
            "best_primary_metric": self.best_primary_metric,
        }

    def _load(self, d: dict) -> None:
        self.model_database = pd.DataFrame(d["model_database"])
        if "best_model_params" in d:
            self.best_model = create_BC(d["best_model_params"])
            self.best_model._load(d["best_model"])
            self.best_primary_metric = d["best_primary_metric"]
        else:
            self.best_model = None
            self.best_primary_metric = float("inf")


class SheetsModelManager(BaseModelManager):
    """SheetsModelManager is the class that tracks model development against a specified
    dataset on Google Sheets. It is responsible for saving parameter settings and metrics.

    Parameters:
        dataset (BaseDataset): The dataset to use for model development.
        metrics (list[Str]): The metrics to track e.g. ROC AUC, Root Mean Squared Error.
        name (str): The name of the Google Sheets to save this to. Optional.
        email (str): The email to share the results to. Optional, Default is
            share to anyone with the link.
    """

    @log_arguments
    def __init__(
        self,
        dataset: BaseDataset,
        metrics: List[str],
        file_path: str = None,
        primary_metric: str = None,
        name: str = "SheetsModelManager",
        email: str = "",
        log=True,
    ):

        import gspread

        if "GOOGLE_CREDENTIALS_FILENAME" in oce.CONFIG:
            self.gc = gspread.oauth(
                credentials_filename=oce.CONFIG["GOOGLE_CREDENTIALS_FILENAME"],
            )
        elif "GOOGLE_SERVICE_ACC_FILENAME" in oce.CONFIG:
            self.gc = gspread.service_account(
                filename=oce.CONFIG["GOOGLE_SERVICE_ACC_FILENAME"]
            )
        else:
            raise Exception(
                "GOOGLE_CREDENTIALS_FILENAME and GOOGLE_SERVICE_ACC_FILENAME not found in CONFIG, please set one of them.\
                https://docs.f.org/en/latest/oauth2.html#oauth-client-id"
            )

        try:
            self.wb = self.gc.open(name)
        except:
            self.wb = self.gc.create(name)

        try:
            self.ws = self.wb.worksheet(dataset.name)
        except:
            self.ws = self.wb.add_worksheet(title=dataset.name, rows=1, cols=1)

        self.email = email

        super().__init__(
            dataset,
            metrics,
            file_path=file_path,
            primary_metric=primary_metric,
            log=False,
        )

    def _save(self) -> dict:
        self.wb.share(self.email, perm_type="anyone", role="writer")
        self.ws.update(
            [self.model_database.columns.values.tolist()]
            + self.model_database.values.tolist()
        )
        self.url = f"https://docs.google.com/spreadsheets/d/{self.wb._properties['id']}/edit#gid=0"
        print(f"Saved to Google Sheets. Please visit {self.url} to view the results.")

    def _load(self, d: dict) -> None:
        self.model_database = pd.DataFrame(self.ws.get_all_records())


class FirebaseModelManager(BaseModelManager):
    """FirebaseModelManager is a ModelManager that saves model parameters
    and performances to a Firebase database.

     A Firebase service account key in oce.CONFIG is required for database access.

    Model information is saved to a collection called 'models' in the database.
    For each document, the following is saved:
        - uid: the user id of the user associated with the model
        - did: the dataset_id of on which the model was trained
        - model_parameters: parameters of the BaseModel oce object
        - model_name
        - model_status
        - fit_time
        - metrics: model training metrics

    Dataset information is saved to a collection called 'datasets' in the database.
    For each document, the following is saved:
        - dataset: map representation of the BaseDataset oce object
        - hashed_dataset: md5 hash of the dataset data
        - uid: the user id of the user associated with the dataset

    Parameters:
        dataset (BaseDataset): The dataset to use for model development.
        metrics (list[Str]): The metrics to track e.g. ROC AUC, Root Mean Squared Error.
        file_path (str): The path to save the model manager to.
        uid (str): The user id associated with the model manager
    """

    @log_arguments
    def __init__(
        self,
        dataset: BaseDataset,
        metrics: List[str],
        uid: str,
        primary_metric: str = None,
        file_path: str = None,
        log=True,
    ):
        import firebase_admin
        from firebase_admin import credentials, firestore

        if "FIREBASE_CREDENTIALS_CERTIFICATE" in oce.CONFIG:
            cred = credentials.Certificate(
                oce.CONFIG["FIREBASE_CREDENTIALS_CERTIFICATE"]
            )
            firebase_admin.initialize_app(cred)
            self.db = firestore.client()
        else:
            raise Exception(
                "Firebase credentials in CONFIG were not found or not properly set. "
            )

        super().__init__(
            dataset,
            metrics,
            file_path=file_path,
            primary_metric=primary_metric,
            log=False,
        )
        self.uid = uid

    def run(self, models: Union[BaseModel, List[BaseModel]]):
        """
        Run the model manager on the given model(s).

        Parameters:
            models (BaseModel or list[BaseModel]): The model(s) to run.
        """
        super().run(models)

        model_database = self.get_model_database()
        dataset = self.dataset
        dataset_data = dataset.data
        hashed_dataset = joblib.hash(dataset_data)
        dataset_params = oce.pretty_params(dataset)
        data_map = {
            "name": dataset.name,
            "file_path": dataset.file_path,
            "structure_col": dataset.structure_col,
            "property_col": dataset.property_col,
            "feature_cols": dataset.feature_cols,
            "hashed_dataset": hashed_dataset,
            "dataset_params": dataset_params,
        }

        # create a new document in datasets
        dataset_ref = self.db.collection("datasets").document()
        dataset_ref.set(
            {
                "uid": self.uid,
                "data": data_map,
            }
        )

        # check if userID is already in users, if not, create a new document in users
        user_ref = self.db.collection("users").document(self.uid)
        if not user_ref.get().exists:
            user_ref.set(
                {
                    "uid": self.uid,
                }
            )

        for model in model_database.to_dict("index").values():
            name: str = model["Model Name"]
            fit_time: float = model["Fitting Time"]
            parameters: dict = model["Model Parameters"]
            metric_results: dict = {
                key: value
                for key, value in model.items()
                if key not in ["Model Name", "Fitting Time", "Model Parameters"]
            }

            model_map = {
                "status": "Trained",
                "name": name,
                "parameters": json.dumps(parameters),
                "metrics": metric_results,
                "fit_time": fit_time,
            }

            # create a new document in models
            model_ref = self.db.collection("models").document()
            model_ref.set(
                {
                    "uid": self.uid,
                    "did": dataset_ref.id,
                    "data": model_map,
                }
            )

    def _save(self) -> dict:
        pass

    def _load(self, d: dict):
        pass



"""
    Setup file for olorenchemengine.
    Use setup.cfg to configure your project.

    This file was generated with PyScaffold 4.1.1.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: https://pyscaffold.org/
"""
import codecs
import os

from setuptools import setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")


if __name__ == "__main__":
    try:
        setup(
            version=get_version("src/olorenchemengine/__version__.py"),
            use_scm_version={"version_scheme": "no-guess-dev"},
        )
    except:  # noqa
        print(
            "\n\nAn error occurred while building the project, "
            "please ensure you have the most updated version of setuptools, "
            "setuptools_scm and wheel with:\n"
            "   pip install -U setuptools setuptools_scm wheel\n\n"
        )
        raise



"""
Model comparisons is a set of visualizations that compare various models or
parameters thereof. This can be used for hyperparameter tuning, benchmarking
visualizations, or model selections.
"""

from olorenchemengine.base_class import *
from olorenchemengine.dataset import *
from olorenchemengine.internal import *
from olorenchemengine.interpret import *
from olorenchemengine.manager import *
from olorenchemengine.representations import *
from olorenchemengine.uncertainty import *
from olorenchemengine.visualizations import *


class VisualizationModelManager_Bar(BaseVisualization):
    """Visualize the database of a model manager in a bar plot, with the model
    parameters and name displayed on hover

    Parameters:
        mm (ModelManager): The model manager to visualize
        top (int): The number of models to display
    """

    @log_arguments
    def __init__(self, mm: BaseModelManager, top=5, log=True, **kwargs):
        self.df = mm.get_model_database()
        self.metrics = mm.metrics
        self.mm = mm
        self.top = top
        super().__init__(log=False)
        self.packages += ["plotly"]

    @property
    def JS_NAME(self) -> str:
        return "ModelManagerBar"

    def get_data(self, metric=None) -> dict:
        """
        Get the data for the visualization

        Parameters:
            metric (str): The metrics to use for the visualization

        Returns:
            dict: The data for the visualization
        """
        df2 = self.df.sort_values(
            by=metric,
            ascending=(True if self.mm.metric_direction[metric] == "lower" else False),
        )
        df2 = df2.iloc[: self.top]
        d = df2.to_dict("l")
        if not metric is None:
            assert (
                metric in self.metrics
            ), f"Designated metric, {metric}, not in model manager metrics: {self.metrics}"
            d["metric"] = metric
        else:
            d["metric"] = self.metrics[0]
        d["title"] = "Bar Plot of Best Models"
        return d

    def render_ipynb(self, metric=None, *args, print_html_js=False, **kwargs) -> str:
        """
        Render the visualization in an ipython notebook

        Parameters:
            metric (str): The metric to use for the visualization
            *args: Additional arguments to pass to the visualization
            print_html_js (bool): Whether to print the javascript code for the visualization
            **kwargs: Additional arguments to pass to the visualization

        Returns:
            str: The html code for the visualization
        """
        d = self.get_data(metric)
        return self.render_ipynb_from_dict(
            d, *args, print_html_js=print_html_js, **kwargs
        )


class ModelOverTime(CompoundScatterPlot):
    """Visualizes the performance of a given model architecture over a length of time.

    This visualization will use a date split and for each date i, will train a model
    on the data on dates 1, ..., i-1, evaluating on date i. Note, any existing splits
    in the dataset will be ignored

    Parameters:
        dataset (BaseDataset): The dataset to use for the visualization
        model (BaseModel): The model to use for the visualization
    """

    @log_arguments
    def __init__(
        self,
        dataset: BaseDataset,
        model: BaseModel,
        *args,
        log=True,
        title="Model Performance over Time",
        yaxis_title="Error",
        xaxis_title="Date",
        **kwargs,
    ):
        self.dataset = dataset
        self.model = model
        self.title = title
        self.yaxis_title = yaxis_title
        self.xaxis_title = xaxis_title

        self.dataset.data = self.dataset.data.sort_values(by=[self.dataset.date_col])
        self.dates = self.dataset.data[self.dataset.date_col].unique()
        results = pd.DataFrame()
        for i in tqdm(range(len(self.dates) - 1)):
            # try:
            train_date = self.dates[i]
            test_date = self.dates[i + 1]
            train = self.dataset.data[
                self.dataset.data[self.dataset.date_col] <= train_date
            ]
            train["split"] = "train"
            test = self.dataset.data[
                self.dataset.data[self.dataset.date_col] == test_date
            ]
            test["split"] = "test"
            DATA = pd.concat([train, test])
            dataset = (
                oce.BaseDataset(
                    data=DATA.to_csv(), structure_col="SMILES", property_col="log AUC"
                )
                + oce.CleanStructures()
            )
            print(f"Train size is {len(train)} and test size is {len(test)}")
            self.model.fit(*dataset.train_dataset)

            test_set = dataset.data[dataset.data["split"] == "test"]
            test_set["pred"] = self.model.predict(test_set["SMILES"])

            results = pd.concat([results, test_set])

            if self.model.setting == "regression":
                results["Error (|Predicted - True|)"] = np.abs(
                    results["pred"] - results[self.dataset.property_col]
                )
                self.yaxis_title = "Error (|Predicted - True|)"
            elif self.model.setting == "classification":
                results["Error (p or 1-p)"] = np.abs(
                    results[self.dataset.property_col]
                    + (-2 * results[self.dataset.property_col] + 1) * results["pred"]
                )
                self.yaxis_title = "Error (p or 1-p)"

        results["X"] = results[self.dataset.date_col]
        if self.model.setting == "regression":
            results["Y"] = results["Error (|Predicted - True|)"]
            self.xaxis_title = xaxis_title
        elif self.model.setting == "classification":
            results["Y"] = results["Error (p or 1-p)"]
            self.yaxis_title = yaxis_title
        results["SMILES"] = results[self.dataset.structure_col]
        super().__init__(
            *args,
            df=results,
            xaxis_type="date",
            title=title,
            xaxis_title=self.xaxis_title,
            yaxis_title=self.yaxis_title,
            log=False,
            **kwargs,
        )

    def get_data(self):
        d = super().get_data()
        window = 10
        quantile = 0.80
        results = list(
            zip(
                pd.to_datetime(self.df["X"]).apply(lambda x: x.timestamp()),
                self.df["Y"],
            )
        )
        scores, residuals = zip(*sorted(results))
        X = pd.Series(scores).rolling(window).mean()
        y = pd.Series(residuals).rolling(window).quantile(quantile)
        X = np.array(X[~np.isnan(X)])
        y = np.array(y[~np.isnan(y)])

        d["trace_update"] = {
            "x": self.df["X"].tolist(),
            "y": y.tolist(),
            "mode": "markers",
            "type": "scatter",
            "marker": {"color": "red"},
        }
        return d



import math

import torch
import torch.nn.functional as F
import torch_sparse
from torch import nn
from torch.nn import LayerNorm, Linear, Parameter, ReLU
from torch_geometric.nn import (
    GCNConv,
    MessagePassing,
    global_add_pool,
    global_max_pool,
    global_mean_pool,
)
from torch_geometric.utils import add_self_loops, degree, softmax
from torch_geometric.utils.num_nodes import maybe_num_nodes
from torch_scatter import scatter, scatter_add
from torch_sparse import SparseTensor, fill_diag, matmul, mul
from torch_sparse import sum as sparsesum

num_atom_type = 119  # including the extra mask tokens
num_chirality_tag = 3

num_bond_type = 5  # including aromatic and self-loop edge
num_bond_direction = 3


class GINEConv(MessagePassing):
    def __init__(self, emb_dim):
        super(GINEConv, self).__init__()
        self.mlp = nn.Sequential(
            nn.Linear(emb_dim, 2 * emb_dim), nn.ReLU(), nn.Linear(2 * emb_dim, emb_dim)
        )
        self.edge_embedding1 = nn.Embedding(num_bond_type, emb_dim)
        self.edge_embedding2 = nn.Embedding(num_bond_direction, emb_dim)

        nn.init.xavier_uniform_(self.edge_embedding1.weight.data)
        nn.init.xavier_uniform_(self.edge_embedding2.weight.data)

    def forward(self, x, edge_index, edge_attr):
        # add self loops in the edge space
        edge_index = add_self_loops(edge_index, num_nodes=x.size(0))[0]

        # add features corresponding to self-loop edges.
        self_loop_attr = torch.zeros(x.size(0), 2)
        self_loop_attr[:, 0] = 4  # bond type for self-loop edge
        self_loop_attr = self_loop_attr.to(edge_attr.device).to(edge_attr.dtype)
        edge_attr = torch.cat((edge_attr, self_loop_attr), dim=0)

        edge_embeddings = self.edge_embedding1(edge_attr[:, 0]) + self.edge_embedding2(
            edge_attr[:, 1]
        )

        return self.propagate(edge_index, x=x, edge_attr=edge_embeddings)

    def message(self, x_j, edge_attr):
        return x_j + edge_attr

    def update(self, aggr_out):
        return self.mlp(aggr_out)


class GINet(nn.Module):
    """
    Args:
        num_layer (int): the number of GNN layers
        emb_dim (int): dimensionality of embeddings
        drop_ratio (float): dropout rate
        gnn_type: gin, gcn, graphsage, gat
    Output:
        node representations
    """

    def __init__(
        self,
        task="classification",
        num_layer=5,
        emb_dim=300,
        feat_dim=512,
        drop_ratio=0,
        pool="mean",
        pred_n_layer=2,
        pred_act="softplus",
    ):
        super(GINet, self).__init__()
        self.num_layer = num_layer
        self.emb_dim = emb_dim
        self.feat_dim = feat_dim
        self.drop_ratio = drop_ratio
        self.task = task

        self.x_embedding1 = nn.Embedding(num_atom_type, emb_dim)
        self.x_embedding2 = nn.Embedding(num_chirality_tag, emb_dim)
        nn.init.xavier_uniform_(self.x_embedding1.weight.data)
        nn.init.xavier_uniform_(self.x_embedding2.weight.data)

        # List of MLPs
        self.gnns = nn.ModuleList()
        for layer in range(num_layer):
            self.gnns.append(GINEConv(emb_dim))

        # List of batchnorms
        self.batch_norms = nn.ModuleList()
        for layer in range(num_layer):
            self.batch_norms.append(nn.BatchNorm1d(emb_dim))

        if pool == "mean":
            self.pool = global_mean_pool
        elif pool == "max":
            self.pool = global_max_pool
        elif pool == "add":
            self.pool = global_add_pool
        self.feat_lin = nn.Linear(self.emb_dim, self.feat_dim)

        if self.task == "classification":
            out_dim = 1
        elif self.task == "regression":
            out_dim = 1

        self.pred_n_layer = max(1, pred_n_layer)

        if pred_act == "relu":
            pred_head = [
                nn.Linear(self.feat_dim, self.feat_dim // 2),
                nn.ReLU(inplace=True),
            ]
            for _ in range(self.pred_n_layer - 1):
                pred_head.extend(
                    [
                        nn.Linear(self.feat_dim // 2, self.feat_dim // 2),
                        nn.ReLU(inplace=True),
                    ]
                )
            pred_head.append(nn.Linear(self.feat_dim // 2, out_dim))
        elif pred_act == "softplus":
            pred_head = [nn.Linear(self.feat_dim, self.feat_dim // 2), nn.Softplus()]
            for _ in range(self.pred_n_layer - 1):
                pred_head.extend(
                    [nn.Linear(self.feat_dim // 2, self.feat_dim // 2), nn.Softplus()]
                )
        else:
            raise ValueError("Undefined activation function")

        pred_head.append(nn.Linear(self.feat_dim // 2, out_dim))
        self.pred_head = nn.Sequential(*pred_head)

    def forward(self, data):
        x = data.x
        edge_index = data.edge_index
        edge_attr = data.edge_attr

        h = self.x_embedding1(x[:, 0]) + self.x_embedding2(x[:, 1])

        for layer in range(self.num_layer):
            h = self.gnns[layer](h, edge_index, edge_attr)
            h = self.batch_norms[layer](h)
            if layer == self.num_layer - 1:
                h = F.dropout(h, self.drop_ratio, training=self.training)
            else:
                h = F.dropout(F.relu(h), self.drop_ratio, training=self.training)

        h = self.pool(h, data.batch)
        h = self.feat_lin(h)

        return h, self.pred_head(h)

    def load_my_state_dict(self, state_dict):
        own_state = self.state_dict()
        for name, param in state_dict.items():
            if name not in own_state:
                continue
            if isinstance(param, nn.parameter.Parameter):
                # backwards compatibility for serialized parameters
                param = param.data
            own_state[name].copy_(param)


def gcn_norm(edge_index, num_nodes=None):
    num_nodes = maybe_num_nodes(edge_index, num_nodes)

    edge_weight = torch.ones((edge_index.size(1),), device=edge_index.device)

    row, col = edge_index[0], edge_index[1]
    deg = scatter_add(edge_weight, col, dim=0, dim_size=num_nodes)
    deg_inv_sqrt = deg.pow_(-0.5)
    deg_inv_sqrt.masked_fill_(deg_inv_sqrt == float("inf"), 0)
    return edge_index, deg_inv_sqrt[row] * edge_weight * deg_inv_sqrt[col]


class GCNConv(MessagePassing):
    def __init__(self, emb_dim, aggr="add"):
        super(GCNConv, self).__init__()
        self.emb_dim = emb_dim
        self.aggr = aggr

        self.weight = Parameter(torch.Tensor(emb_dim, emb_dim))
        self.bias = Parameter(torch.Tensor(emb_dim))
        self.reset_parameters()

        self.edge_embedding1 = nn.Embedding(num_bond_type, 1)
        self.edge_embedding2 = nn.Embedding(num_bond_direction, 1)

        nn.init.xavier_uniform_(self.edge_embedding1.weight.data)
        nn.init.xavier_uniform_(self.edge_embedding2.weight.data)

    def reset_parameters(self):
        # glorot(self.weight)
        # zeros(self.bias)
        stdv = math.sqrt(6.0 / (self.weight.size(-2) + self.weight.size(-1)))
        self.weight.data.uniform_(-stdv, stdv)
        self.bias.data.fill_(0)

    def forward(self, x, edge_index, edge_attr):
        # add self loops in the edge space
        edge_index = add_self_loops(edge_index, num_nodes=x.size(0))[0]

        # add features corresponding to self-loop edges.
        self_loop_attr = torch.zeros(x.size(0), 2)
        self_loop_attr[:, 0] = 4  # bond type for self-loop edge
        self_loop_attr = self_loop_attr.to(edge_attr.device).to(edge_attr.dtype)
        edge_attr = torch.cat((edge_attr, self_loop_attr), dim=0)

        edge_embeddings = self.edge_embedding1(edge_attr[:, 0]) + self.edge_embedding2(
            edge_attr[:, 1]
        )

        edge_index, __ = gcn_norm(edge_index)

        x = x @ self.weight

        # propagate_type: (x: Tensor, edge_attr: OptTensor)
        out = self.propagate(edge_index, x=x, edge_attr=edge_embeddings, size=None)

        if self.bias is not None:
            out += self.bias

        return out

    def message(self, x_j, edge_attr):
        # return x_j if edge_attr is None else edge_attr.view(-1, 1) * x_j
        return x_j if edge_attr is None else edge_attr + x_j

    def message_and_aggregate(self, adj_t, x):
        return torch_sparse.matmul(adj_t, x, reduce=self.aggr)


class GCN(nn.Module):
    def __init__(
        self,
        task="classification",
        num_layer=5,
        emb_dim=300,
        feat_dim=256,
        drop_ratio=0,
        pool="mean",
    ):
        super(GCN, self).__init__()
        self.num_layer = num_layer
        self.emb_dim = emb_dim
        self.feat_dim = feat_dim
        self.drop_ratio = drop_ratio
        self.task = task

        if self.num_layer < 2:
            raise ValueError("Number of GNN layers must be greater than 1.")

        self.x_embedding1 = nn.Embedding(num_atom_type, emb_dim)
        self.x_embedding2 = nn.Embedding(num_chirality_tag, emb_dim)

        nn.init.xavier_uniform_(self.x_embedding1.weight.data)
        nn.init.xavier_uniform_(self.x_embedding2.weight.data)

        # List of MLPs
        self.gnns = nn.ModuleList()
        for layer in range(num_layer):
            self.gnns.append(GCNConv(emb_dim, aggr="add"))

        # List of batchnorms
        self.batch_norms = nn.ModuleList()
        for layer in range(num_layer):
            self.batch_norms.append(nn.BatchNorm1d(emb_dim))

        if pool == "mean":
            self.pool = global_mean_pool
        elif pool == "add":
            self.pool = global_add_pool
        elif pool == "max":
            self.pool = global_max_pool
        else:
            raise ValueError("Not defined pooling!")

        self.feat_lin = nn.Linear(self.emb_dim, self.feat_dim)

        if self.task == "classification":
            self.pred_head = nn.Sequential(
                nn.Linear(self.feat_dim, self.feat_dim // 2),
                nn.Softplus(),
                nn.Linear(self.feat_dim // 2, 1),
            )
        elif self.task == "regression":
            self.pred_head = nn.Sequential(
                nn.Linear(self.feat_dim, self.feat_dim // 2),
                nn.Softplus(),
                nn.Linear(self.feat_dim // 2, 1),
            )

    def forward(self, data):
        x = data.x
        edge_index = data.edge_index
        edge_attr = data.edge_attr

        h = self.x_embedding1(x[:, 0]) + self.x_embedding2(x[:, 1])

        for layer in range(self.num_layer):
            h = self.gnns[layer](h, edge_index, edge_attr)
            h = self.batch_norms[layer](h)
            if layer == self.num_layer - 1:
                h = F.dropout(h, self.drop_ratio, training=self.training)
            else:
                h = F.dropout(F.relu(h), self.drop_ratio, training=self.training)

        h = self.pool(h, data.batch)
        h = self.feat_lin(h)

        return h, self.pred_head(h)

    def load_my_state_dict(self, state_dict):
        own_state = self.state_dict()
        for name, param in state_dict.items():
            if name not in own_state:
                continue
            if isinstance(param, nn.parameter.Parameter):
                # backwards compatibility for serialized parameters
                param = param.data
            own_state[name].copy_(param)
